from pathlib import Path
import struct
import ctypes
from typing import List
import pyart
from .c_wrapper.function_alias import aliases
from .c_wrapper.conversions import list_to_array, array_to_list


def parse_boundary_file(boundary_path: str):
    """ extracts two lists, for x and y points, of boundary file """
    path_to_boundary_file = Path(boundary_path)

    data = path_to_boundary_file.read_bytes()
    iterations = int((len(data) - 20) / 8)

    i = int.from_bytes(data[:5], byteorder="little", signed=True)
    # print("header", ints)

    # print("\n{:<8} {:<8} {:<8} {:<8} {:<8}".format(
    #     "i", "start", "end", "x", "y"))

    x_points = []
    y_points = []

    for i in range(iterations):
        start = 20 + i * 8
        end = start + 8
        x, y = struct.unpack("ff", data[start:end])
        x_points.append(int(x))
        y_points.append(int(y))
        # print("{:<8} {:<8} {:<8} {:<8} {:<8}".format(i, start, end, x, y))

    # print()
    return x_points, y_points

# 0  10
# 10 0
# 0  -10
# -10 0


def get_boundary_mask_from_file(radar: pyart.core.Radar, boundary_path: str) -> List[List[bool]]:
    x_points, y_points = parse_boundary_file(boundary_path)
    return get_boundary_mask_from_list(radar, x_points, y_points)


def get_boundary_mask_from_list(radar: pyart.core.Radar, x_points: List[int], y_points: List[int]) -> List[List[bool]]:

    # x_points, y_points = ([20, 20, -20, -20], [20, -20, -20, 20])

    # boundary mask function returns void
    aliases["get_boundary_mask"].restype = None

    # define the boundary mask function's argument types
    aliases["get_boundary_mask"].argtypes = (
        ctypes.POINTER(ctypes.c_long),
        ctypes.POINTER(ctypes.c_long),
        ctypes.c_int,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.c_int,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_float,
        ctypes.c_float,
        ctypes.POINTER(ctypes.c_bool),
    )

    # extract these lists from the radar's data
    radar_azimuth_list = radar.azimuth['data'].tolist()

    #  =====Scan modes=====
    # define              CAL 0
    # define              PPI 1
    # define              COP 2
    # define              RHI 3
    # define              VER 4
    # define              TAR 5
    # define              MAN 6
    # define              IDL 7
    # define              SUR 8
    # define              AIR 9
    # define              HOR 10

    #  ====Radar types=====
    # define           GROUND 0
    # define         AIR_FORE 1
    # define          AIR_AFT 2
    # define         AIR_TAIL 3
    # define           AIR_LF 4
    # define             SHIP 5
    # define         AIR_NOSE 6
    # define        SATELLITE 7
    # define     LIDAR_MOVING 8
    # define      LIDAR_FIXED 9

    radar_scan_mode = 1  # default to PPI
    radar_scan_type = 0  # default to GROUND

    if radar.scan_type == 'rhi':
        radar_scan_mode = 3

    if 'platform_type' in radar.metadata:
        if radar.metadata['platform_type'] == 'aircraft_tail':
            radar_scan_type = 3
        elif radar.metadata['platform_type'] == 'fixed':
            radar_scan_type = 9

    boundary_list_of_lists = []


    for i in range(radar.nrays):
        boundary = list_to_array([True] * radar.ngates, ctypes.c_bool)
        x_array = list_to_array(x_points, ctypes.c_long)
        y_array = list_to_array(y_points, ctypes.c_long)

        aliases["get_boundary_mask"](
            x_array,  # x_points
            y_array,  # y_points
            ctypes.c_int(len(x_points)),  # npoints

            ctypes.c_float(radar.latitude['data'].tolist()[0]),  # radar_origin_latitude
            ctypes.c_float(radar.longitude['data'].tolist()[0]),  # radar_origin_longitude
            ctypes.c_float(radar.altitude['data'].tolist()[0]),  # radar_origin_altitude

            ctypes.c_float(0),  # boundary_origin_tilt
            ctypes.c_float(radar.latitude['data'].tolist()[0]),  # boundary_origin_latitude
            ctypes.c_float(radar.longitude['data'].tolist()[0]),  # boundary_origin_longitude
            ctypes.c_float(radar.altitude['data'].tolist()[0]),  # boundary_origin_altitude

            ctypes.c_int(radar.ngates),  # nGates
            ctypes.c_float(radar.range['meters_between_gates'] / 1000),  # gateSize
            ctypes.c_float(radar.range['meters_to_center_of_first_gate'] / 1000),  # distanceToCellInMeters
            ctypes.c_float(radar_azimuth_list[i]),  # azimuth
            ctypes.c_int(radar_scan_mode),  # radar_scan_mode
            ctypes.c_int(radar_scan_type),  # radar_type
            ctypes.c_float(0),  # tilt_angle
            ctypes.c_float(0),  # rotation_angle ----- not used by code
            boundary  # boundary_mask_out
        )

        boundary_list_of_lists.append(array_to_list(boundary, radar.ngates))

    return boundary_list_of_lists



import ctypes
import pyart
import numpy as np
from ..c_wrapper.run_solo import run_solo_function
from ..c_wrapper import DataPair, masked_op
from ..c_wrapper.function_alias import aliases


se_despeckle = aliases['despeckle']


def despeckle_ray(input_list_data, bad, a_speckle, dgi_clip_gate=None, boundary_mask=None):
    """
        Performs a despeckle operation on a list of data (a single ray)

        Args:
            input_list_data: A list containing float data,
            bad: A float that represents a missing/invalid data point,
            a_speckle: An integer that determines the number of contiguous good data considered a speckle,
            (optional) dgi_clip_gate: An integer determines the end of the ray (default: length of input_list),
            (optional) boundary_mask: this is the masked region bool list where the function will perform its operation (default: all True, so operation performed on entire region).

        Returns:
            Numpy masked array: Contains an array of data, mask, and fill_value of results.

    """

    args = {
        "data": DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data),
        "newData": DataPair.DataTypeValue(np.ctypeslib.ndpointer(ctypes.c_float, flags="C_CONTIGUOUS"), None),
        "nGates": DataPair.DataTypeValue(ctypes.c_size_t, None),
        "bad": DataPair.DataTypeValue(ctypes.c_float, bad),
        "a_speckle": DataPair.DataTypeValue(ctypes.c_size_t, a_speckle),
        "dgi_clip_gate": DataPair.DataTypeValue(ctypes.c_size_t, dgi_clip_gate),
        "boundary_mask": DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), boundary_mask),
    }

    return run_solo_function(se_despeckle, args)


def despeckle_masked(masked_array, a_speckle, boundary_masks=None):
    return masked_op.masked_func_iterable(despeckle_ray, masked_array, {'boundary_mask': boundary_masks}, {'a_speckle': a_speckle})


def despeckle_field(radar: pyart.core.Radar, field: str, new_field: str, a_speckle: int, boundary_masks=None, sweep=0):

    with masked_op.SweepManager(radar, sweep, field, new_field) as sm:
        sm.new_masked_array = despeckle_masked(sm.radar_sweep_data, a_speckle, boundary_masks)



import os
import sys
from contextlib import contextmanager

# https://stackoverflow.com/questions/5081657/how-do-i-prevent-a-c-shared-library-to-print-on-stdout-in-python/17954769#17954769

@contextmanager
def stdout_redirected(to=os.devnull):
    '''
    import os

    with stdout_redirected(to=filename):
        print("from Python")
        os.system("echo non-Python applications are also supported")
    '''
    fd = sys.stdout.fileno()

    ##### assert that Python and C stdio write using the same file descriptor
    ####assert libc.fileno(ctypes.c_void_p.in_dll(libc, "stdout")) == fd == 1

    def _redirect_stdout(to):
        sys.stdout.close() # + implicit flush()
        os.dup2(to.fileno(), fd) # fd writes to 'to' file
        sys.stdout = os.fdopen(fd, 'w') # Python writes to fd

    with os.fdopen(os.dup(fd), 'w') as old_stdout:
        with open(to, 'w') as file:
            _redirect_stdout(to=file)
        try:
            yield # allow code to be run with the redirected stdout
        finally:
            _redirect_stdout(to=old_stdout) # restore stdout.
                                            # buffering and flags such as
                                            # CLOEXEC may be different



import ctypes
import numpy as np
from ..c_wrapper.run_solo import run_solo_function

from ..c_wrapper import DataPair, masked_op
from ..c_wrapper.function_alias import aliases

se_remove_storm_motion = aliases['remove_storm_motion']

def remove_storm_motion_ray(input_list_data, bad, wind, speed, dgi_dd_rotation_angle, dgi_dd_elevation_angle, dgi_clip_gate=None, boundary_mask=None):
    """ 
        Performs a <TODO> operation on a list of data.
        
        Args:
            input_list: A list containing float data,
            bad: A float that represents a missing/invalid data point,
            wind: <TODO>
            speed: <TODO>
            dgi_dd_rotation_angle: <TODO>
            dgi_dd_elevation_angle: <TODO>
            (optional) dgi_clip_gate: An integer determines the end of the ray (default: length of input_list)
            (optional) boundary_mask: Defines region over which operations will be done. (default: all True).

        Returns:
          Numpy masked array: Contains an array of data, mask, and fill_value of results.

        Throws:
          ValueError: if input_list and input_boundary_mask are not equal in size,
    """

    args = {
        "wind" : DataPair.DataTypeValue(ctypes.c_float, wind),
        "speed" : DataPair.DataTypeValue(ctypes.c_float, speed),
        "dgi_dd_rotation_angle" : DataPair.DataTypeValue(ctypes.c_float, dgi_dd_rotation_angle),
        "dgi_dd_elevation_angle" : DataPair.DataTypeValue(ctypes.c_float, dgi_dd_elevation_angle),
        "data" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data),
        "newData" : DataPair.DataTypeValue(np.ctypeslib.ndpointer(ctypes.c_float, flags="C_CONTIGUOUS"), None),
        "nGates" : DataPair.DataTypeValue(ctypes.c_size_t, None),
        "bad" : DataPair.DataTypeValue(ctypes.c_float, bad),
        "dgi_clip_gate" : DataPair.DataTypeValue(ctypes.c_size_t, dgi_clip_gate),
        "boundary_mask" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), boundary_mask),
    }

    return run_solo_function(se_remove_storm_motion, args)


def remove_storm_motion_masked(masked_array, wind, speed, dgi_dd_rotation_angle, dgi_dd_elevation_angle, boundary_masks=None):
    """ 
        Performs a <TODO> operation on a numpy masked array
        
        Args:
            masked_array: A numpy masked array data structure,
            wind: <TODO>
            speed: <TODO>
            dgi_dd_rotation_angle: <TODO>
            dgi_dd_elevation_angle: <TODO>

        Returns:
            Numpy masked array

        Throws:
            ModuleNotFoundError: if numpy is not installed
            AttributeError: if masked_array arg is not a numpy masked array.
    """


    return masked_op.masked_func(remove_storm_motion_ray, masked_array, wind, speed, dgi_dd_rotation_angle, dgi_dd_elevation_angle, boundary_masks = boundary_masks)



import ctypes

def array_to_list(input_array, size):
    """ converts ctypes array to Python list """
    return [input_array[i] for i in range(size)]


def list_to_array(py_list, type_in_c):
    """ convert Python list to ctypes array """

    a = (type_in_c * len(py_list))()
    a[:] = py_list

    return a



"""
Python interface for Solo II
"""

from .solo_functions.solo_despeckle import *
from .solo_functions.solo_ring_zap import *
from .solo_functions.solo_threshold import *
from .solo_functions.solo_flag_glitches import *
from .solo_functions.solo_flag_freckles import *
from .solo_functions.solo_forced_unfolding import *
from .solo_functions.solo_unfold_first_good_gate import unfold_first_good_gate, unfold_first_good_gate_masked
from .solo_functions.solo_unfold_local_wind import *
from .solo_functions.solo_radial_shear import *
from .solo_functions.solo_rain_rate import *
from .solo_functions.solo_remove_ac_motion import remove_ac_motion_ray, remove_ac_motion_masked
from .solo_functions.solo_remove_storm_motion import remove_storm_motion_ray, remove_storm_motion_masked
# from .solo_functions.solo_fix_vortex_vels import fix_vortex_vels, fix_vortex_vels_masked
from .solo_functions.solo_merge_fields import *
from .solo_functions.flags.solo_assign_value import assign_value, assign_value_masked
from .solo_functions.flags.solo_assert_bad_flags import assert_bad_flags, assert_bad_flags_masked
from .solo_functions.flags.solo_bad_flags_logic import bad_flags_logic, bad_flags_logic_masked
from .solo_functions.flags.solo_copy_bad_flags import copy_bad_flags, copy_bad_flags_masked
from .solo_functions.flags.solo_flagged_add import flagged_add, flagged_add_masked
from .solo_functions.flags.solo_set_bad_flags import set_bad_flags, set_bad_flags_masked
from .solo_functions.flags.solo_clear_bad_flags import clear_bad_flags
from .boundary import *
from .enums import Where, Logical


'''
    This module is a wrapper for run_solo. Allows use of run_solo with 2D masked arrays rather than 1D arrays.
'''

import numpy as np

def masked_func(func, masked_array, *args, boundary_masks=None, second_masked_array=None, usesBadFlags=False):
    """
        Performs a solo function on a 2D masked array.

        Args:
            masked_array: A numpy masked array data structure with input data (usually this contains info on a single field),
            args: A list of args, listed in order, as required by the solo function
            (optional) boundary_masks: A boolean list designating which region to perform the operation. (Default = all True, so entire region is operated on.)
            (optional) second_masked_array: Some solo functions do operations on two fields. If so, assign this parameter to the masked array of that other field.
            (optional) usesBadFlags: Some solo functions do operations on masks rather than data. If so, set to true.

        Returns:
            Numpy masked array

        Throws:
            AttributeError: if masked_array arg is not a numpy masked array.
    """

    missing = masked_array.fill_value
    mask = masked_array.mask.tolist()
    data_list = masked_array.tolist(missing)

    if second_masked_array is not None:
        second_missing = second_masked_array.fill_value
        second_data_list = second_masked_array.tolist(second_missing)

    if not mask:
        mask = [False] * len(data_list)
    if not boundary_masks:
        boundary_masks = [[True] * len(data_list[0])] * len(data_list)

    # initialize lists with data/masks. These will become lists of lists
    output_data = []
    output_mask = []

    # iterate through each ray
    for i in range(len(data_list)):
        # for i in range(len(data_list)):
        input_data = data_list[i]  # gates
        input_mask = mask[i]  # mask for gates

        if second_masked_array is not None:  # if using second masked array, obtain gates for that
            second_input_data = second_data_list[i]
            # in pysolo, second parameter of functions always designates data list for secondary masked list, for functions that have secondary masked arrays.
            func_ma = func(input_data, second_input_data, missing, *args, boundary_mask=boundary_masks[i])
        elif usesBadFlags:
            func_ma = func(input_data, missing, *args, input_mask, boundary_mask=boundary_masks[i])
        else:
            func_ma = func(input_data, missing, *args, boundary_mask=boundary_masks[i])

        output_data.append(np.ma.getdata(func_ma))  # add data to output list of lists
        output_mask.append(np.ma.getdata(func_ma.mask))  # add mask as well

    # convert the list of lists, to a 2D masked array
    output_masked_array = np.ma.masked_array(data=output_data, mask=output_mask, fill_value=missing)
    return output_masked_array


def masked_func_iterable(func, masked_array, iterables, statics):

    # initialize lists with data/masks. These will become lists of lists
    output_data = []
    output_mask = []

    params_dict = {}

    missing = masked_array.fill_value
    params_dict['bad'] = missing

    for key, value in statics.items():
        params_dict[key] = value

    data_list = masked_array.tolist(missing)

    # iterate through each ray
    for i in range(len(data_list)):
        params_dict['input_list_data'] = data_list[i]
        for key, value in iterables.items():
            params_dict[key] = None if value is None else value[i]

        func_ma = func(**params_dict)

        output_data.append(np.ma.getdata(func_ma))  # add data to output list of lists
        output_mask.append(np.ma.getdata(func_ma.mask))  # add mask as well

    # convert the list of lists, to a 2D masked array
    output_masked_array = np.ma.masked_array(data=output_data, mask=output_mask, fill_value=missing)
    return output_masked_array


class SweepManager:

    def __init__(self, radar, sweep, field, new_field) -> None:
        self.radar = radar
        self.sweep = sweep
        self.field = field
        self.new_field = new_field
        self.new_field_data: np.ma.array = radar.fields[field]['data'].copy()
        self.radar_sweep_data = radar.get_field(sweep, field)
        self.new_masked_array = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):

        a = self.radar.get_slice(self.sweep)

        for ray in range(a.start, a.stop):
            self.new_field_data[ray] = self.new_masked_array[ray]

        self.radar.add_field_like(self.field, self.new_field, self.new_field_data, replace_existing=True)



import ctypes
import pyart
import numpy as np

from ..c_wrapper.run_solo import run_solo_function
from ..c_wrapper import DataPair, masked_op
from ..c_wrapper.function_alias import aliases

se_unfold_local_wind = aliases['BB_unfold_local_wind']


def unfold_local_wind_ray(
        input_list_data, bad, nyquist_velocity, dds_radd_eff_unamb_vel, azimuth_angle_degrees, elevation_angle_degrees, ew_wind, ns_wind, ud_wind, max_pos_folds, max_neg_folds, ngates_averaged,
        dgi_clip_gate=None, boundary_mask=None):
    """
        Performs a <TODO>

        Args:
            input_list: A list containing float data,
            bad: A float that represents a missing/invalid data point,
            nyquist_velocity: <TODO>,
            dds_radd_eff_unamb_vel: <TODO>,
            azimuth_angle_degrees: <TODO>,
            elevation_angle_degrees: <TODO>,
            ew_wind: <TODO>,
            ns_wind: <TODO>,
            ud_wind: <TODO>,
            max_pos_folds: <TODO>,
            max_neg_folds: <TODO>,
            ngates_averaged: <TODO>,
            last_good_v0: <TODO>,
            (optional) dgi_clip_gate: An integer determines the end of the ray (default: length of input_list)
            (optional) boundary_mask: Defines region over which operations will be done. (default: all True).

        Returns:
          Numpy masked array: Contains an array of data, mask, and fill_value of results.

        Throws:
          ValueError: if input_list and input_boundary_mask are not equal in size,
    """

    args = {
        "data": DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data),
        "newData" : DataPair.DataTypeValue(np.ctypeslib.ndpointer(ctypes.c_float, flags="C_CONTIGUOUS"), None),
        "nGates": DataPair.DataTypeValue(ctypes.c_size_t, None),
        "nyquist_velocity": DataPair.DataTypeValue(ctypes.c_float, nyquist_velocity),
        "dds_radd_eff_unamb_vel": DataPair.DataTypeValue(ctypes.c_float, dds_radd_eff_unamb_vel),
        "azimuth_angle_degrees": DataPair.DataTypeValue(ctypes.c_float, azimuth_angle_degrees),
        "elevation_angle_degrees": DataPair.DataTypeValue(ctypes.c_float, elevation_angle_degrees),
        "ew_wind": DataPair.DataTypeValue(ctypes.c_float, ew_wind),
        "ns_wind": DataPair.DataTypeValue(ctypes.c_float, ns_wind),
        "ud_wind": DataPair.DataTypeValue(ctypes.c_float, ud_wind),
        "max_pos_folds": DataPair.DataTypeValue(ctypes.c_int, max_pos_folds),
        "max_neg_folds": DataPair.DataTypeValue(ctypes.c_int, max_neg_folds),
        "ngates_averaged": DataPair.DataTypeValue(ctypes.c_size_t, ngates_averaged),
        "bad": DataPair.DataTypeValue(ctypes.c_float, bad),
        "dgi_clip_gate": DataPair.DataTypeValue(ctypes.c_size_t, dgi_clip_gate),
        "boundary_mask": DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), boundary_mask),
    }

    return run_solo_function(se_unfold_local_wind, args)


def unfold_local_wind_masked(
        masked_array, nyquist_velocity, dds_radd_eff_unamb_vel, azimuth_angle_degrees, elevation_angle_degrees, ew_wind, ns_wind, ud_wind, max_pos_folds, max_neg_folds, ngates_averaged,
        boundary_masks=None):
    """
        Performs a <TODO> on a numpy masked array

        Args:
            masked_array: A numpy masked array data structure,
            nyquist_velocity: <TODO>
            dds_radd_eff_unamb_vel: <TODO>
            max_pos_folds: <TODO>
            max_neg_folds: <TODO>
            ngates_averaged: <TODO>
            last_good_v0: <TODO>

        Returns:
            Numpy masked array

        Throws:
            ModuleNotFoundError: if numpy is not installed
            AttributeError: if masked_array arg is not a numpy masked array.
    """

    return masked_op.masked_func_iterable(
        unfold_local_wind_ray,
        masked_array,
        {
            'boundary_mask': boundary_masks,
            'azimuth_angle_degrees': azimuth_angle_degrees,
            'elevation_angle_degrees': elevation_angle_degrees
        },
        {
            'nyquist_velocity': nyquist_velocity,
            'dds_radd_eff_unamb_vel': dds_radd_eff_unamb_vel,
            'ew_wind': ew_wind,
            'ns_wind': ns_wind,
            'ud_wind': ud_wind,
            'max_pos_folds': max_pos_folds,
            'max_neg_folds': max_neg_folds,
            'ngates_averaged': ngates_averaged
        }
    )


def unfold_local_wind_fields(radar: pyart.core.Radar, field: str, new_field: str, ew_wind, ns_wind, ud_wind, max_pos_folds, max_neg_folds, ngates_averaged, boundary_masks=None, sweep=0):

    with masked_op.SweepManager(radar, sweep, field, new_field) as sm:
        print(sm)
        nyquist_velocity = sm.radar.get_nyquist_vel(sm.sweep)
        dds_radd_eff_unamb_vel = nyquist_velocity
        azimuth_angle_degrees = list(sm.radar.get_azimuth(sm.sweep))
        elevation_angle_degrees = list(sm.radar.get_elevation(sm.sweep))

        sm.new_masked_array = unfold_local_wind_masked(sm.radar_sweep_data, nyquist_velocity, dds_radd_eff_unamb_vel,
                                                    azimuth_angle_degrees, elevation_angle_degrees, ew_wind, ns_wind, ud_wind, max_pos_folds, max_neg_folds, ngates_averaged, boundary_masks)
